#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'


# import standard library

# import interfaces
from pyams_form.interfaces.form import IFormHelp
from pyams_security.interfaces import IAdminAuthenticationPlugin, ISecurityManager
from pyams_skin.interfaces.viewlet import IToolbarAddingMenu
from pyams_skin.layer import IPyAMSLayer
from pyams_utils.interfaces import VIEW_SYSTEM_PERMISSION, MANAGE_SYSTEM_PERMISSION
from z3c.form.interfaces import DISPLAY_MODE, IDataExtractedEvent
from zope.component.interfaces import ISite

# import packages
from pyams_form.form import AJAXAddForm, AJAXEditForm
from pyams_form.help import FormHelp
from pyams_pagelet.pagelet import pagelet_config
from pyams_security.plugin.admin import AdminAuthenticationPlugin
from pyams_security.zmi.utility import SecurityManagerPluginsTable
from pyams_skin.viewlet.toolbar import ToolbarMenuItem
from pyams_utils.adapter import adapter_config
from pyams_utils.registry import query_utility
from pyams_viewlet.viewlet import viewlet_config
from pyams_zmi.form import AdminDialogEditForm, AdminDialogAddForm
from pyams_zmi.layer import IAdminLayer
from pyramid.events import subscriber
from pyramid.url import resource_url
from pyramid.view import view_config
from z3c.form import field
from zope.interface import Invalid

from pyams_security import _


@viewlet_config(name='add-admin-authentication.menu', context=ISite, layer=IAdminLayer,
                view=SecurityManagerPluginsTable, manager=IToolbarAddingMenu,
                permission=MANAGE_SYSTEM_PERMISSION, weight=1)
class AdminAuthenticationAddMenu(ToolbarMenuItem):
    """Admin authentication add menu"""

    label = _("Add admin authentication...")
    label_css_class = 'fa fa-fw fa-support'
    url = 'add-admin-authentication.html'
    modal_target = True


@pagelet_config(name='add-admin-authentication.html', context=ISite, layer=IPyAMSLayer,
                permission=MANAGE_SYSTEM_PERMISSION)
class AdminAuthenticationAddForm(AdminDialogAddForm):
    """Admin authentication plug-in add form"""

    title = _("System security manager")
    legend = _("Add administration authentication plug-in")
    icon_css_class = 'fa fa-fw fa-support'

    fields = field.Fields(IAdminAuthenticationPlugin).omit('__name__', '__parent__')
    autocomplete = 'off'
    ajax_handler = 'add-admin-authentication.json'
    edit_permission = MANAGE_SYSTEM_PERMISSION

    def create(self, data):
        return AdminAuthenticationPlugin()

    def add(self, plugin):
        context = query_utility(ISecurityManager)
        context[plugin.prefix] = plugin

    def nextURL(self):
        return resource_url(self.context, self.request, 'security-manager.html')


@subscriber(IDataExtractedEvent, form_selector=AdminAuthenticationAddForm)
def handle_new_admin_folder_data_extraction(event):
    """Handle new admin authentication plug-in form data extraction"""
    manager = query_utility(ISecurityManager)
    prefix = event.data.get('prefix')
    if prefix in manager:
        event.form.widgets.errors += (Invalid(_("Specified prefix is already used!")),)


@view_config(name='add-admin-authentication.json', context=ISite, request_type=IPyAMSLayer,
             permission=MANAGE_SYSTEM_PERMISSION, renderer='json', xhr=True)
class AdminAuthenticationAJAXAddForm(AJAXAddForm, AdminAuthenticationAddForm):
    """Admin authentication plug-in add form, AJAX handler"""


@pagelet_config(name='properties.html', context=IAdminAuthenticationPlugin, layer=IPyAMSLayer,
                permission=VIEW_SYSTEM_PERMISSION)
class AdminAuthenticationEditForm(AdminDialogEditForm):
    """Admin authentication plug-in edit form"""

    @property
    def title(self):
        return self.context.title

    legend = _("Edit administration authentication plug-in")
    icon_css_class = 'fa fa-fw fa-support'

    fields = field.Fields(IAdminAuthenticationPlugin).omit('__name__', '__parent__')
    autocomplete = 'off'
    ajax_handler = 'properties.json'
    edit_permission = MANAGE_SYSTEM_PERMISSION

    def updateWidgets(self, prefix=None):
        super(AdminAuthenticationEditForm, self).updateWidgets()
        self.widgets['prefix'].mode = DISPLAY_MODE


@view_config(name='properties.json', context=IAdminAuthenticationPlugin, request_type=IPyAMSLayer,
             permission=MANAGE_SYSTEM_PERMISSION, renderer='json', xhr=True)
class AdminAuthenticationAJAXEditForm(AJAXEditForm, AdminAuthenticationEditForm):
    """Admin authentication plug-in edit form, AJAX handler"""


@adapter_config(context=(IAdminAuthenticationPlugin, IAdminLayer, AdminAuthenticationEditForm), provides=IFormHelp)
class AdminAuthenticationHelpAdapter(FormHelp):
    """Admin authentication edit form help adapter"""

    header = _("WARNING")
    status = 'danger'
    message = _("""Before disabling plug-in, please verify that you have other administration access!""")
    message_format = 'rest'
