#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'


# import standard library
from datetime import datetime

# import interfaces
from pyams_pagelet.interfaces import PageletCreatedEvent
from pyams_security.interfaces import IUsersFolderPlugin, ISecurityManager, ILocalUser, IUserRegistrationInfo
from pyams_skin.interfaces.viewlet import IToolbarAddingMenu, IWidgetTitleViewletManager
from pyams_skin.interfaces import IPageHeader
from pyams_skin.layer import IPyAMSLayer
from pyams_utils.interfaces import VIEW_SYSTEM_PERMISSION, MANAGE_SYSTEM_PERMISSION
from pyams_zmi.layer import IAdminLayer
from z3c.form.interfaces import DISPLAY_MODE, IDataExtractedEvent
from z3c.table.interfaces import IColumn
from zope.component.interfaces import ISite
from zope.dublincore.interfaces import IZopeDublinCore

# import packages
from pyams_form.form import AJAXAddForm, AJAXEditForm
from pyams_form.search import SearchView, SearchResultsView
from pyams_pagelet.pagelet import pagelet_config
from pyams_security.plugin.userfolder import UsersFolder, User
from pyams_security.zmi.utility import SecurityManagerPluginsTable
from pyams_skin.page import DefaultPageHeaderAdapter
from pyams_skin.table import I18nColumn
from pyams_skin.viewlet.toolbar import ToolbarMenuItem, ToolbarAction
from pyams_utils.adapter import adapter_config
from pyams_utils.date import format_datetime
from pyams_utils.registry import query_utility
from pyams_utils.url import absolute_url
from pyams_viewlet.viewlet import viewlet_config
from pyams_zmi.form import AdminDialogAddForm, AdminDialogEditForm
from pyams_zmi.view import AdminView
from pyramid.events import subscriber
from pyramid.view import view_config
from z3c.form import field
from z3c.table.column import GetAttrColumn
from zope.interface import Invalid

from pyams_security import _


#
# Users folder views
#

@viewlet_config(name='add-users-folder.menu', context=ISite, layer=IAdminLayer,
                view=SecurityManagerPluginsTable, manager=IToolbarAddingMenu,
                permission=MANAGE_SYSTEM_PERMISSION, weight=10)
class UsersFolderAddMenu(ToolbarMenuItem):
    """Local users folder add menu"""

    label = _("Add local users folder...")
    label_css_class = 'fa fa-fw fa-user'
    url = 'add-users-folder.html'
    modal_target = True


@pagelet_config(name='add-users-folder.html', context=ISite, layer=IPyAMSLayer,
                permission=MANAGE_SYSTEM_PERMISSION)
class UsersFolderAddForm(AdminDialogAddForm):
    """Users folder plug-in add form"""

    title = _("System security manager")
    legend = _("Add local users folder plug-in")
    icon_css_class = 'fa fa-fw fa-user'

    fields = field.Fields(IUsersFolderPlugin).omit('__name__', '__parent__')
    ajax_handler = 'add-users-folder.json'
    edit_permission = MANAGE_SYSTEM_PERMISSION

    def create(self, data):
        return UsersFolder()

    def add(self, plugin):
        context = query_utility(ISecurityManager)
        context[plugin.prefix] = plugin

    def nextURL(self):
        return absolute_url(self.context, self.request, 'security-manager.html')


@subscriber(IDataExtractedEvent, form_selector=UsersFolderAddForm)
def handle_new_user_folder_data_extraction(event):
    """Handle new user folder form data extraction"""
    manager = query_utility(ISecurityManager)
    prefix = event.data.get('prefix')
    if prefix in manager:
        event.form.widgets.errors += (Invalid(_("Specified prefix is already used!")),)


@view_config(name='add-users-folder.json', context=ISite, request_type=IPyAMSLayer,
             permission=MANAGE_SYSTEM_PERMISSION, renderer='json', xhr=True)
class UsersFolderAJAXAddForm(AJAXAddForm, UsersFolderAddForm):
    """users folder plug-in add form, AJAX handler"""


@pagelet_config(name='properties.html', context=IUsersFolderPlugin, layer=IPyAMSLayer,
                permission=VIEW_SYSTEM_PERMISSION)
class UsersFolderEditForm(AdminDialogEditForm):
    """Users folder plug-in edit form"""

    @property
    def title(self):
        return self.context.title

    legend = _("Edit local users folder plug-in properties")
    icon_css_class = 'fa fa-fw fa-user'

    fields = field.Fields(IUsersFolderPlugin).omit('__name__', '__parent__')
    ajax_handler = 'properties.json'
    edit_permission = MANAGE_SYSTEM_PERMISSION

    def updateWidgets(self, prefix=None):
        super(UsersFolderEditForm, self).updateWidgets()
        self.widgets['prefix'].mode = DISPLAY_MODE


@view_config(name='properties.json', context=IUsersFolderPlugin, request_type=IPyAMSLayer,
             permission=MANAGE_SYSTEM_PERMISSION, renderer='json', xhr=True)
class UsersFolderAJAXEditForm(AJAXEditForm, UsersFolderEditForm):
    """Users folder plug-in edit form, AJAX handler"""


#
# Users folder search views
#

@pagelet_config(name='search.html', context=IUsersFolderPlugin, layer=IPyAMSLayer, permission=VIEW_SYSTEM_PERMISSION)
class UsersFolderSearchView(AdminView, SearchView):
    """Users folder search view"""
    
    def __init__(self, context, request):
        super(UsersFolderSearchView, self).__init__(context, request)


@adapter_config(context=(IUsersFolderPlugin, IAdminLayer, UsersFolderSearchView), provides=IPageHeader)
class UsersFolderSearchViewHeaderAdapter(DefaultPageHeaderAdapter):
    """Users folder search view header adapter"""

    back_url = '#security-manager.html'
    icon_class = 'fa fa-fw fa-user'
    title = _("Security manager")

    @property
    def subtitle(self):
        return self.context.title


@view_config(name='search-results.html', context=IUsersFolderPlugin, request_type=IPyAMSLayer,
             permission=VIEW_SYSTEM_PERMISSION)
class UsersFolderSearchResultsView(AdminView, SearchResultsView):
    """Users folder search results view table"""

    id = 'users_folder_search_table'
    title = _("Search results")
    cssClasses = {'table': 'table table-bordered table-striped table-hover table-tight datatable'}

    def __init__(self, context, request):
        super(UsersFolderSearchResultsView, self).__init__(context, request)
        request.registry.notify(PageletCreatedEvent(self))


@adapter_config(name='login', context=(IUsersFolderPlugin, IAdminLayer, UsersFolderSearchResultsView),
                provides=IColumn)
class LoginColumn(I18nColumn, GetAttrColumn):
    """Users login column"""

    _header = _("Login")
    attrName = 'login'
    weight = 5


@adapter_config(name='name', context=(IUsersFolderPlugin, IAdminLayer, UsersFolderSearchResultsView),
                provides=IColumn)
class NameColumn(I18nColumn, GetAttrColumn):
    """Users name column"""

    _header = _("Name")
    attrName = 'title'
    weight = 10


@adapter_config(name='email', context=(IUsersFolderPlugin, IAdminLayer, UsersFolderSearchResultsView),
                provides=IColumn)
class EmailColumn(I18nColumn, GetAttrColumn):
    """Users email column"""

    _header = _("E-mail address")
    attrName = 'email'
    weight = 20


@adapter_config(name='registration_date', context=(IUsersFolderPlugin, IAdminLayer, UsersFolderSearchResultsView),
                provides=IColumn)
class RegistrationDateColumn(I18nColumn, GetAttrColumn):
    """Users registration date column"""

    _header = _("Registration date")
    weight = 30

    def getValue(self, obj):
        dc = IZopeDublinCore(obj, None)
        if dc is not None:
            return format_datetime(dc.created)
        else:
            return '--'


@adapter_config(name='activation_date', context=(IUsersFolderPlugin, IAdminLayer, UsersFolderSearchResultsView),
                provides=IColumn)
class ConfirmationDateColumn(I18nColumn, GetAttrColumn):
    """Users activation date column"""

    _header = _("Activation date")
    weight = 40

    def getValue(self, obj):
        if obj.activation_date:
            return format_datetime(obj.activation_date)
        else:
            return '--'


#
# Users views
#

@viewlet_config(name='users-folder.toolbar.adding', context=IUsersFolderPlugin,
                view=UsersFolderSearchView.search_form_factory, layer=IAdminLayer,
                manager=IWidgetTitleViewletManager, permission=MANAGE_SYSTEM_PERMISSION)
class LocalUserAddAction(ToolbarAction):
    """Users folder adding action"""

    label = _("Add user")
    url = 'add-user.html'
    modal_target = True


@pagelet_config(name='add-user.html', context=IUsersFolderPlugin, layer=IPyAMSLayer,
                permission=MANAGE_SYSTEM_PERMISSION)
class LocalUserAddForm(AdminDialogAddForm):
    """Local user add form"""

    @property
    def title(self):
        return self.context.title

    legend = _("Add new local user")
    icon_css_class = 'fa fa-fw fa-user'
    label_css_class = 'control-label col-md-4'
    input_css_class = 'col-md-8'

    fields = field.Fields(IUserRegistrationInfo).select('login', 'email', 'firstname', 'lastname', 'company_name') + \
        field.Fields(ILocalUser).select('password_manager') + \
        field.Fields(IUserRegistrationInfo).select('password', 'confirmed_password') + \
        field.Fields(ILocalUser).select('wait_confirmation')

    autocomplete = 'off'
    ajax_handler = 'add-user.json'
    edit_permission = MANAGE_SYSTEM_PERMISSION

    def updateWidgets(self, prefix=None):
        super(LocalUserAddForm, self).updateWidgets()
        self.widgets['password'].input_css_class = 'col-md-4'
        self.widgets['confirmed_password'].input_css_class = 'col-md-4'

    def create(self, data):
        return User()

    def update_content(self, user, data):
        user.login = data.get('login')
        user.email = data.get('email')
        user.firstname = data.get('firstname')
        user.lastname = data.get('lastname')
        user.company_name = data.get('company_name')
        user.password_manager = data.get('password_manager')
        user.password = data.get('password')
        user.self_registered = False
        if data.get('wait_confirmation'):
            user.generate_secret()
        else:
            user.activation_date = datetime.utcnow()
            user.activated = True
        user.wait_confirmation = data.get('wait_confirmation')

    def add(self, user):
        self.context[user.login] = user


@subscriber(IDataExtractedEvent, form_selector=LocalUserAddForm)
def handle_new_user_data_extraction(event):
    """Handle new user form data extraction"""
    data = event.data
    if not data.get('login'):
        data['login'] = data['email']
    folder = event.form.context
    if not folder.check_login(data.get('login')):
        event.form.widgets.errors += (Invalid(_("Specified login can't be used!")),)


@view_config(name='add-user.json', context=IUsersFolderPlugin, request_type=IPyAMSLayer,
             permission=MANAGE_SYSTEM_PERMISSION, renderer='json', xhr=True)
class LocalUserAJAXAddForm(AJAXAddForm, LocalUserAddForm):
    """Local user add form, AJAX view"""

    def get_ajax_output(self, changes):
        translate = self.request.localizer.translate
        return {'status': 'success',
                'message': translate(_("User was created successfully"))}


@pagelet_config(name='properties.html', context=ILocalUser, layer=IPyAMSLayer, permission=VIEW_SYSTEM_PERMISSION)
class LocalUserEditForm(AdminDialogEditForm):
    """Local user edit form"""

    @property
    def title(self):
        return self.context.title

    legend = _("Edit user properties")
    icon_css_class = 'fa fa-fw fa-user'
    label_css_class = 'control-label col-md-4'
    input_css_class = 'col-md-8'

    fields = field.Fields(ILocalUser).select('login', 'email', 'firstname', 'lastname', 'company_name',
                                             'self_registered', 'activation_hash', 'activation_date')

    ajax_handler = 'properties.json'
    edit_permission = MANAGE_SYSTEM_PERMISSION

    def updateWidgets(self, prefix=None):
        super(LocalUserEditForm, self).updateWidgets()
        self.widgets['self_registered'].mode = DISPLAY_MODE
        self.widgets['activation_hash'].mode = DISPLAY_MODE
        self.widgets['activation_date'].mode = DISPLAY_MODE


@view_config(name='properties.json', context=ILocalUser, request_type=IPyAMSLayer,
             permission=MANAGE_SYSTEM_PERMISSION, renderer='json', xhr=True)
class LocalUserAJAXEditForm(AJAXEditForm, LocalUserEditForm):
    """Local user edit form, AJAX view"""
